# Heimtools.pm - Cern Heimdal (kerberos5) tools: routines to supplement some additional tools
#
# Bernard Antoine 11/2004
#
# $Header: /afs/cern.ch/project/afs/dev/security/RCS/Heimtools.pm,v 1.107 2013/05/07 14:43:44 qba Exp $
# $Pgm:  /afs/cern.ch/project/afs/perlmodules
# $Man:  /afs/cern.ch/project/afs/doc
# $Html: /afs/cern.ch/project/afs/doc

package Heimtools;

use strict;
use Exporter;
use vars qw (
              @ISA @EXPORT $VERSION
            );
@ISA = qw(Exporter);
@EXPORT = qw ( 
              setpag authenticate get_users block_unblock
              set_pwstring new_passwd set_expstring expire_passwd userid_exclude Krun Knote
              __Build_ticket _read_ans __Do_auth __Do_token __Do_kinit _locate _refmt
              $legal_username_pattern $legal_nodename_pattern $legal_volname_pattern
             );
$VERSION = 1.03;

use vars qw( %created_on %created_by %modified_on %modified_by %pw_expires %pw_expired %expires %expired %is_user
             $legal_username_pattern $legal_nodename_pattern $legal_volname_pattern
             $logger $kadmin $kinit $klog @Kprinc $short_life $long_life $K_debug $tmpfs @Kident
             $has_AFS_pag
            );


# same constants appear also in Vos.pm
$legal_username_pattern = '[a-z][\w-]+';   # a-z followed by a-zA-Z_0-9-   ( . excluded )
$legal_nodename_pattern = '[a-z][\w-]+';   # ...             a-zA-z_0-9-
$legal_volname_pattern  = '[a-z][\w-.]+';  # ...             a-zA-z_0-9-.  ( . included )

$short_life  = 5;
$long_life   = 365;
$K_debug = 0;
$logger = "/usr/bin/logger";
$kadmin = "/usr/heimdal/sbin/kadmin";
if ( -d "/usr/heimdal/bin/" ) {
    $kinit  = "/usr/heimdal/bin/kinit --afslog=no";
} elsif ( -e "/usr/kerberos/bin/kinit" ) {
    $kinit  = "/usr/kerberos/bin/kinit";
} elsif ( -e "/usr/bin/kinit" ) {
    $kinit  = "/usr/bin/kinit";
}
$klog   = "/usr/bin/aklog";
@Kprinc = qw( -p ___afs -K /etc/krb5.keytab.afs );     # watch $Kprinc[...] if this format changes

# variables
$tmpfs = ( -d "/ttmp" ) ? "/ttmp" : "/tmp";
@Kident = @Kprinc; 
if ( -r "/var/heimdal/.is_master" && $> == 0 ) { @Kident = ("-l"); }

sub setpag {
    chomp(my $a = `/afs/usr/local/etc/afsutil -p`);
    &llog("pag now 1: $a") if $K_debug;
    if ($a > 0)  { $has_AFS_pag = 1; }
    my($dir) = @_;
    if ($dir) {
        $ENV{'KRB5CCNAME'} = "FILE:$dir/krb5_tkfile";
    } elsif ($ENV{'KRB5CCNAME'}) {
        ;
    } elsif ($a > 0) {
        $ENV{'KRB5CCNAME'} = "FILE:$tmpfs/krb5_$a";
    } else {
        use File::Temp qw / tempfile /;
        my ($fh, $fn) = &tempfile( "/tmp/krb5_XXXXXXXX");
        $ENV{'KRB5CCNAME'} = "FILE:$fn";
        END:
        close $fh;
        unlink $fn;
    }
}


sub authenticate {
    # there are three assumptions here:
    # 1-  if /usr/heimdal/bin exists, ALL of heimdal is installed,
    #     else ALL af MIT kerberos5 is installed.
    # 2-  /etc/hosts contains in order: ip_address long_hostname short_hostname [logname]
    require "/afs/cern.ch/project/afs/etc/enqdeq.pm";
    my($principal, $keytab) = @_;
    my($klist, $arg, @stat, $must, $brincipal, $mit);
    my %life = ('krbtgt', 18000, 'kadmin/admin', 1200);   # resp. 5 h & 20 mn
    while (-e "/tmp/WAITFORKDC") { system "/bin/sleep 5"; }
    use Time::Local;

    if ( $principal eq "root" ) {
        chomp( my $host = `hostname` );
        my($lhost, $shost) = gethostbyname($host);
        $shost =~ s/([^\s]+)(?:\s.+)*/$1/;
        $principal = "host/$lhost";
        $brincipal = "rcmd.$shost";
    } elsif ( ! $principal ) {
        $brincipal = $principal = $Kprinc[1];
        $keytab =                 $Kprinc[3];
    } else {
        ($brincipal = $principal) =~ s!/!.!;
    }

    my $myrealm = &get_realm;
    &setpag() ; # no-OP is KRB5CCNAME already set, else set default directory for tickets

    if ( -e "/usr/kerberos/bin/klist" )  {
       $klist = "/usr/kerberos/bin/klist";
       $mit = 1;
       use Time::Local;
    } elsif (-e "/usr/heimdal/bin/klist") {
       $klist = "/usr/heimdal/bin/klist -v";
       use Date::Parse;
    } elsif (-e "/usr/bin/klist" ) {
	$klist = "/usr/bin/klist";
    }

    # we could be only partly root... and klist would fail in the krb4 part
    my $suid;
    if ($> == 0 and $< != 0) { # running in suid mode?
        $suid = $<;    # keep orig
        $<    = $>;    # become real root
    }
    @stat = `$klist 2>/dev/null`;
    $< = $suid if $suid;  # revert

    $must = 1;
    my $trigger = 0;
    my $now = time;
    foreach (@stat) {
       if ( /^$/ ) { ; }
       elsif ( 
                  m!\Qrincipal: $principal\E\@\Q$myrealm\E!
              || 
                  m!\QPrincipal: $brincipal\E\@\Q$myrealm\E!
              ) { $must = $must-1; }
       elsif (/>>>Expired<<</) { $must = $must+1 ; }
       elsif ($trigger && /^Server:\s/) { $trigger = 0; }
       elsif ( /^Server:\s+krbtgt/ ) { $trigger = 1; }
       elsif ($mit && / krbtgt[\/\.]/) {
          my($m,$d,$y,$H,$M,$S) = m!^\S+\s+\S+\s+(\d+)/(\d+)/(\d+)\s+(\d+):(\d+):(\d+)\s+krb!;
          my $date = timelocal($S,$M,$H,$d,$m-1,$y);
          my $now = time;
          &llog("$S,$M,$H,$d,$m,$y -> $date, time=$now") if $K_debug;
          unless ($date > ($now+7200)) { $must = $must+1 ; }
       }
       elsif ($trigger && /^End time:\s+(.+)(\s+\(expired\))?$/) {
          $trigger = 0;
          my $date = str2time($1);
          my $now = time;
          &llog("$1 -> $date, time=$now") if $K_debug;
          unless ($date > ($now+7200)) { $must = $must+1 ; }
       }
    }
    my $rc;
    if ($must) {
       &llog("must=$must") if $K_debug;

       $arg = ($keytab) ? "-t $keytab" : "" ;
       my $i = 5;
       while (1) {
          my $suid;
          if ($> == 0 and $< != 0) { # running in suid mode?
              $suid = $<;    # keep orig
              $<    = $>;    # become real root
          }
          enq('kinit_lock');
          my $Q =`$kinit -k $arg $principal 2>&1`;
          $< = $suid if $suid;  # revert
          $rc = $?;
          deq('kinit_lock');
          if ($rc) {
             next if ( $Q =~ /ASN\.1 value too large/ );     # known case, retry "forever"
             &llog("command '$kinit -k $arg $principal' returned\n$Q");
             next if $i--;                                   # else retry a few times
             &llog("tkt file: $ENV{'KRB5CCNAME'}\n");
             &llog("principal=$principal, brincipal=$brincipal, realm=$myrealm\n");
             &llog("@stat");
             die "cannot re-authenticate as $principal. rc from kinit is $rc\n";
          }
          last;
       }
       continue {
          sleep 2;
       }
    } 
    enq('kinit_lock');
    if ($has_AFS_pag) { ;
                      }
                else  { system($klog,"-setpag"); 
                        $rc=$?;
                        $has_AFS_pag=1;
                        chomp(my $a = `/afs/usr/local/etc/afsutil -p`);
                        &llog("pag now 2: $a") if $K_debug;
                      }
    system($klog);
    $rc=$?;
    &llog("aklog rc=$rc") if $K_debug or $rc;
    deq('kinit_lock');
    !$rc;

}


sub get_users {

#
# we are supposed to be already root, on an AFS server
#
    my ($remote) = (@_, "");
    my $kadmin_cmd = "$kadmin -l dump";
    my $myrealm = &get_realm;
    if ( $remote || -r "/var/heimdal" ) {
       while (-e "/tmp/WAITFORKDC") { system "/bin/sleep 5"; }
       open(ALL, "$kadmin_cmd |") || die "Can not execute $kadmin_cmd: $!\n";
    } else {
       my $arc = "/afs/usr/local/etc/arc";
       my @kdc = &get_kdc;
       open(ALL, "$arc -l -p -h $kdc[0] kas list |")
           || die "Cannot open arc connection to $kdc[0]\n";
    }
    
    my @x = localtime;
    $x[5] += 1900;
    $x[4] += 1;
    my $now = sprintf( "%04d%02d%02d%02d%02d%02d", $x[5], $x[4], $x[3], $x[2], $x[1], $x[0] );

    while ($_ = <ALL>) {
       my ($principal, $keys, $created, $modified, $valid_start, $valid_end,
        $pw_expires, $maxlife, $maxrenew, $flags, $generation) = split(/\s+/, $_);

       if ($remote eq 'R') {     #  called through arc kas list
          print "$principal - $created $modified $valid_start $valid_end " .
            "$pw_expires $maxlife $maxrenew $flags $generation\n";
          next;
       }

       my ($user, $realm) = $principal =~ /(.*)\@(.*)/;
       next if $realm ne $myrealm;

       ($is_user{$user},$created_on{$user},$modified_on{$user},$pw_expires{$user},$expires{$user}) = (0,0,0,0,0);
       ($created_by{$user},$modified_by{$user}) = ("","");
       ($pw_expired{$user},$expired{$user}) = (0,0);

       $modified = $created if $modified eq "-";
       ($created_on{$user},  $created_by{$user})  = $created  =~ /(.*):(.*)/ unless $created  eq '-';
       ($modified_on{$user}, $modified_by{$user}) = $modified =~ /(.*):(.*)/ unless $modified eq '-';
    
       $is_user{$user}    = 1;
       $pw_expires{$user} = $pw_expires          unless $pw_expires eq '-';
       $pw_expired{$user} = ($pw_expires < $now) unless $pw_expires eq '-';
       $expires{$user}    = $valid_end           unless $valid_end  eq '-';
       $expired{$user}    = ($valid_end < $now)  unless $valid_end  eq '-';
    }
    close(ALL);

}


##
##
## kadmin related functions. Can only run where a KDC runs
##
## note: caller is supposed to be root already, and be authenticated
##
##

#------------------------------------------------------------
# block_unblock
# Role : block/unblock an AFS account
#
#------------------------------------------------------------
sub block_unblock {
    my($userid,$to_block,$reason)= (@_,0);
    if ($K_debug) {
       print "would run &block_unblock($userid,$to_block),$reason\n";
       return 0;
    }

    my ($is_active, $status, $rc);

    # this subroutine will check the validity of the call and do the home directory block/unblock
    # we make it an "anonymous" subroutine so all external variables ($is_active, $status,...) are
    #    bound at call time. This avoids the "will not stay shared" warning message.
    my $check_blocked = sub {
        if ($status =~ /does not exist/) {
            print $status;
            return -1;
        }
        my $rc;
        my $initial = substr($userid,0,1);
        my $home = "/afs/cern.ch/user/$initial/$userid";
        my $save = "/afs/cern.ch/ukokot/$initial/$userid";
        my $fs  = -x '/usr/afs/bin/fs'  ? '/usr/afs/bin/fs'  : '/usr/bin/fs';
        my $pts = -x '/usr/afs/bin/pts' ? '/usr/afs/bin/pts' : '/usr/bin/pts';

        if ($to_block) {
            unless ($is_active) {
                print "account is already blocked\n" if $reason;
                print $status if $reason;
                return -2;
            } 
            $rc = 1;
            return $rc if $userid =~ /[.\/:\@]/;

            # changing status fron unblock to block: save ACLs & reset them to minimum
            my $acls = `$fs la $home`; $acls =~ s/\n/ /g;
            my $pattern = " is ";
            if ($acls =~ /Normal rights:/)   { $pattern .= "Normal rights: (.+)" }
                                      else   { $pattern .= "()"; }
            if ($acls =~ /Negative rights:/) { $pattern .= "Negative rights: (.+)" }
                                        else { $pattern .= "()"; }
            my ($plus, $minus) = $acls =~ /$pattern$/;

            # if kerberos blocking failed, but we already saved acl, we should better avoid saving
            # them again, as they are now stripped off.
            if ($plus eq '  system:administrators rlidwka ' && ! $minus) {
               print "Home directory of $userid already blocked\n";
               return -2;
            }

            if ($plus) {
                unless( open (F, ">$save/.plus_acl") ) {
                   print "Cannot open $save/.plus_acl for write\n";
                   return -1;
                }
                print F $plus;
                close(F);
            } else { unlink "$save/.plus_acl"; }
            if ($minus) {
                unless( open (F, ">$save/.minus_acl") ) {
                   print "Cannot open $save/.minus_acl for write\n";
                   return -1;
                }
                print F $minus;
                close(F);
            } else { unlink "$save/.minus_acl"; }
            system "$fs sa $home system:administrators all -clear";

        } else {              # unblocking a user
            my $plus = "";
            if ($is_active) {
                print "account had not been blocked\n" if $reason;
                print $status if $reason;
		return -3;

                ## if however ACL had been cleaned by a missed block, we should better restore them
                #my $acls = `$fs la $home`; $acls =~ s/\n/ /g;
                #my $pattern = " is Normal rights: (.+)";
                #($plus) = $acls =~ /$pattern$/;
                #unless ($plus eq '  system:administrators rlidwka ') {
                #   return -1;
                #}
                #print "Home directory of $userid found blocked\n";
            }
            $rc = 0;
            return $rc if $userid =~ /[.\/:\@]/;

            # changing status from block to unblock: reset ACLs from saved info
            if ( -r "$save/.plus_acl") {
                unless( open (F, "$save/.plus_acl") ) {
                   print "Cannot open $save/.plus_acl for read\n";
                   return -1;
                }
                (my $plus = <F>) =~ s/^\s+//;
                close(F);
                my %ACLs = split /\s+/, $plus;
                foreach (keys %ACLs) {
                   system "$fs sa $home $_ $ACLs{$_}" 
                           if `$pts exa $_`;  # fails if account deleted
                }
            }
            if ( -r "$save/.minus_acl") {
                unless( open (F, "$save/.minus_acl") ) {
                   print "Cannot open $save/.minus_acl for read\n";
                   return -1;
                }
                (my $minus = <F>) =~ s/^\s+//;
                close(F);
                my %ACLs = split /\s+/, $minus;
                foreach (keys %ACLs) {
                   system "$fs sa $home $_ $ACLs{$_} -negative" 
                           if `$pts exa $_`;  # fails if account deleted
                }
            }

            return -1 if $plus;   # $plus set only if $is_active, but HOME found blocked

        }
        print "$reason\n" if $reason;
        return $rc;
    };

    &userid_exclude($userid);

# arne 29 APR 2013 : don't alk to Heimdal any longer 
#    while (-e "/tmp/WAITFORKDC") { system "/bin/sleep 5"; }
#    $status = `$kadmin -l get $userid`;    # get current status
#    return -1 unless $status;              # error message sent
#    $is_active = $status =~ /Principal expires: never/;

    $is_active = $to_block;
    my $x = &$check_blocked;

    # -1 consistency check failed
    # -2 block a blocked user
    # -3 unblock an unblocked user
    return($x) if (($x == -1) or ($x == -2) or ($x == -3));

#    my $exp = ($x == 1) ?  "now" : "never";

#    my @cmd = ($kadmin, @Kident, "modify", "--expiration-time=$exp", $userid);
#    $rc = &Krun(@cmd);
    
    # examine entry
#    if ($reason) {
#        my @cmd = ($kadmin, @Kident, "get", $userid);
#        &Krun(@cmd);
#    }
    return $rc;

}


#------------------------------------------------------------
# set_pwstring
# Role : prepare the pw= argument for kadmin command
#
#    receives principal and $pw =
#                 -r      to create a random password
#                 -       to read password from STDIN
#                 <string>
#    returns pwstring ready for kadmin command
#
#------------------------------------------------------------
sub set_pwstring {

    my ($user, $pw, $pfx) = (@_, "", "");
    $pfx .= " " if $pfx;
    if ($pw eq "-r") {
        return "--random-key";
    }
    elsif ($pw eq '-') {        # comes from stdin
        chomp($pw=<STDIN>);
    }
    if ($pw eq '')          { die "${pfx}missing password\n"; }
    if (length($pw) > 64)   { die "${pfx}Invalid password detected, longer than 64 characters\n"; }
##  if ($pw=~/([;`\&\|])/)  { die "${pfx}Invalid password detected, contains '$1'\n"; }

    my $pwstring = "--password=$pw";
    if ($user =~ /^rcmd\.(.+)$/) {
        my $KAS_CMD = -e "/usr/sbin/kas" ? "/usr/sbin/kas" : "/usr/afs/bin/kas" ;
        my($key) = `$KAS_CMD stringtokey $pw` =~ /key=\'(.*)\'\./;   # \nnn\nnn\nnn
        eval "\$key = \"$key\"";                                     # -> 0xabcdefgh
        $pwstring = "--key=" . unpack('H16',$key);                   # -> string in printable hexa
    }
    return $pwstring;
}


#------------------------------------------------------------
# new_passwd
# Role : force a new password for an account
# Example :
# new_passwd("olemoign", "--password=xyz") -> set password "xyz" for account olemoign
# new_passwd("olemoign", "--random-key")   -> set random password for account olemoign
#
# can also provide a lifetime as third argument
#------------------------------------------------------------
sub new_passwd {
    my ($userid,$pwstring,$delay) = @_;
    
    print("Forcing new password for account $userid\n");
    my @cmd = ($kadmin, @Kident, "passwd", $pwstring, $userid);
    &Krun(@cmd,1000);
    &Knote($userid);
    if ($delay) { expire_passwd($userid, $delay); }
}


sub set_expstring {
    my ($delay) = @_;
    if ($delay ne "never") {
      $delay = 0 unless $delay =~ /^[-+]?\d+$/;
      my @time = localtime( $delay*86400 + time() );
      $delay = sprintf("%04d-%02d-%02d", 1900 + $time[5], $time[4]+1, $time[3]);   # Year-month-day
    }
    return "$delay";
}


sub Krun {
    my (@cmd) = @_;
    my $err;
    require "/afs/cern.ch/project/afs/etc/enqdeq.pm";
    while (-e "/tmp/WAITFORKDC") { system "/bin/sleep 5"; }
    if ($cmd[-1] =~ /^\d+$/) {
       $err = pop @cmd;
    } else {
       $err = 0;
    }

    if ($K_debug) {
       print "would run (" . join(", ",@cmd) . ")\n";
    } else {

       foreach my $i (0 ... $#cmd) {
          next unless $cmd[$i] =~ /^([\w-]+)=(.+)$/;
          next if $2 =~ /^[\w-]+$/;
          my ($key, $value) = ($1, $2);
          $value =~ s/'/'"'"'/g;   # end quoted string. dblquote single quote, start quoted string
          $value = "'$value'";     #  add surrounding single quotes
          $value=~ s/''//g;        #    fix pb with adjacent single quotes
          $cmd[$i] = "$key=$value";
       }
 
       my $suid;
       if ($> == 0 and $< != 0) { # running in suid mode?
           $suid = $<;    # keep orig
           $<    = $>;    # become real root
       }
 
       enq('kadmin_lock');
       my $res = `@cmd 2>&1`;
       my $rc = $?;
       deq('kadmin_lock');
 
       $< = $suid if $suid;  # revert
 
       if ($err) {
          die $err+$rc . " $res" if $res;
          return $res;
       } else {
          print $res if $res;
          return $rc;
       }
    }
}


sub Knote {
   my ($userid) = @_;
   return if $userid =~ /[\/\.]/;
   my $initial = substr($userid,0,1);
   my $save = "/afs/cern.ch/ukokot/$initial/$userid";
   return unless -d $save;
   my $from = "";
   for (my $i=1; $i<10; $i++) {
      my @a = caller($i);
      unless (@a) {
         $from = "craspi";
         last;
      }
      if ($a[3] =~ /^main::(.+)$/) {
         $from = $1;
         last;
      }
   }
   return unless $from;
   if ($K_debug) {
      print "would put $from in $save/.pwchanged\n";
   } else {
      open F, ">$save/.pwchanged" or return;
      print F $from;
      close F;
   }
}

sub expire_passwd {
    my ($userid, $delay) = @_;
    my @cmd = ($kadmin, @Kident, "modify", "--pw-expiration-time=" . &set_expstring($delay), $userid);
    &Krun(@cmd,2000);
    &Knote($userid);
}


## routines to create a ticket for a given user , and to install it on a client machine
## on a AFS server, run __Build_ticket.

## on the client, _read_ans reads the result, and separates $token, $krb5ticket and an
##   optional additional string
## then __Do_kinit to install the krb5 ticket, and refresh it

## two additionnal subroutines are part of the package:  _refmt & _locate


#------------------------------------------------------------------------------------------
#
# creates composite ticket & token for user at remote host
#   - locally extract user key into a temporary keytab
#   - kinit that user using the temp keytab
#   - if bneeded, create token out of ticket
#   - return either token or ticket 
#
#------------------------------------------------------------------------------------------
sub __Build_ticket {
    ## this routine relies on a Log and a _refmt routine to exist

    # Dual Mode: Determine flavour of this machine
    my $master_certificate = "/etc/active_directory_master_certificate.pem";

    $ENV{KRB5_CONFIG} = "/etc/krb5.conf.windows.pkinit";

    my ($user, $lifetime, $want_KRB, $pagset, $debug) = (@_, 0, 0);
#   require "/afs/cern.ch/project/afs/etc/enqdeq.pm";

    my $caller;
    if ( defined $main::proc ) { $caller = $main::proc; }
    elsif ( $ENV{SCPRINCIPAL} ) { ($caller = $ENV{SCPRINCIPAL}) =~ s![\@/]!.!g; }
    else { $caller = "Heimtools"; }
    &llog( "caller: $caller") if $K_debug;
    my $srvtab = "$tmpfs/${caller}_K$$";
    my $GetToken = _locate("GetToken", "/usr/sue/etc", "/afs/usr/local/etc");
    my $result = "";

## some trick to stop processing requests when updating the kdc's
    while (-e "/tmp/WAITFORKDC") { system "/bin/sleep 5"; }

#       the lines marked ### are necessary when not using kadmin -l
###     setpag('$tktdir');           ### to possibly reuse the tickets
###     &authenticate("register/admin");    ### renew credentials if necessary

    &setpag() unless $pagset;            # build KRB5CCNAME from AFS pag
    my($k5file) = $ENV{'KRB5CCNAME'} =~ /FILE:(.*)/;
    &main::Log("Kerberos environment: k5:$k5file") if $debug;
    my $lk5;
    while (1) {           # normally runs once
       my $R=$<;$<=$>;    #   be 100%  root again, else the 2>&1 will prevent kinit from running
#      enq('kadmin_lock');

       my $errmsg;

       $errmsg = `$kinit -C FILE:$master_certificate  -l ${lifetime}s -A -f $user 2>&1`;


       my $retry_kinit = 2000;

       while (($errmsg =~ "Password incorrect") && ($retry_kinit > 0))
       {
           $errmsg = `$kinit -C FILE:$master_certificate  -l ${lifetime}s -A -f $user 2>&1`;
           &main::Log("Kinit Attempt $retry_kinit : Kinit result $errmsg") if $debug;
           $retry_kinit--;
       }

       my $rc = $?;
#      deq('kadmin_lock');
       $<=$R;
       $lk5 = -s "$k5file";
       &main::Log("rc from kinit: $rc. result krb5: $lk5 bytes") if $debug;
       if ($rc) {
         if ( $errmsg =~ /ASN\.1 value too large/ ) { sleep(2); next; }
         &main::Print("No ticket created for user $user. rc from kinit: $rc, message is: $errmsg");
         if ($errmsg =~ /has expired while getting/) {     # known and expected case
             unlink($k5file); unlink("$srvtab");
         }
         return "";
       } else {
         $errmsg = "" if $errmsg =~ /ticket lifetime is/;  # other NOTICE: messages are sent
         &main::Print("kinit returned \"$errmsg\"") if $errmsg;
       }
       last;
    }

    my $k5tkt = "";
    my $token = "";

    if ($want_KRB) {
        open(F, "$k5file") || &main::Die("cannot open $k5file\n$!");
        sysread(F, $k5tkt, $lk5);
        close(F);
        &main::Log("got a k5 ticket. result:" . _refmt($k5tkt)) if $debug;
    }
    else {                 # user only wants AFS token
       $klog = "/usr/bin/aklog";
       unless ( $has_AFS_pag) {
       # black magic starts
       # aklog -setpag creates a pag and a token, but not good enough for GetToken
       # so we'll run another normal aklog later.
         system($klog,"-setpag");
         $has_AFS_pag=1;
         chomp(my $a = `/afs/usr/local/etc/afsutil -p`);
         &llog("pag now 2: $a") if $debug;
       # $token = `$GetToken`;
       }
       system($klog);
       $token = `$GetToken`;
    }

    $result = sprintf ('%06d%s%06d%s',
            length($token) , $token ,
            length($k5tkt) , $k5tkt );

    if ( -s $k5file ) {
        unlink($k5file); unlink("$srvtab");
    }
    &main::Log("got token & ticket. result:" . _refmt($result)) if $debug;

    return $result;
}

sub _read_ans {
#   read composite answer from argument FileHandler
#   returns the obtained string
#   relies on external $main::Debug variable, and external Log function
#
#   answer format is:  <len> token <len> [ ticket ] data
#   <len> fields are 6 bytes long and contain length in character.
#   if there is no ticket, the second length field IS 000000
#
#   returns token, ticket & data

    my($from_S) = @_;
    # read answer, then splice it        
    my $offset = 0;        
    my ($l,$k);
    my $reply = "";
    while ($l = sysread($from_S, $reply, 4096, $offset)) {
        $offset += $l;        
        &main::Log( "  read $l bytes...") if $main::Debug;
    }        
    close $from_S;
    &main::Log("received " . length($reply) . " bytes") if $main::Debug;

#   extract the token
    my $token = "";
    &main::Die("Invalid answer from server: null reply") if 0 == length($reply);
    $l = substr($reply,$[,6);
    &main::Die("Invalid answer from server: $reply") unless $l =~ /\d+/;
    $k = $l+6;
    &main::Die("Invalid answer from server: k=$k ") if $k > length($reply);
    $token = substr($reply,$[+6,$l) if $l;
    &main::Log("token" . _refmt($token)) if $main::Debug;
    $reply = substr($reply,$[+$k);      # now points to ticket data

#   extracts K5 ticket if provided.
    my $krb5ticket = "";
    &main::Die("Invalid answer from server: null reply") if 0 == length($reply);
    $l = substr($reply,$[,6);         # could be 0 !
    &main::Die("Invalid answer from server: $reply") unless $l =~ /\d+/;
    $k = $l+6;
    &main::Die("Invalid answer from server: k=$k ") if $k > length($reply);
    $krb5ticket = substr($reply,$[+6,$l) if $l;
    &main::Log("krb5 ticket" . _refmt($krb5ticket)) if $krb5ticket && $main::Debug;
    $reply = substr($reply,$[+$k);      # now points to data part

    # Dual Mode: If there is a flag to indicate the flavour on the TGT, determine what it is set to, and return it.
    my $flavour="";
    if (length($reply) >= 7 && substr($reply,$[,6) =~ /\d+/ ) {
       $l = substr($reply,$[,6);         # could be 0 !
       &main::Die("Invalid answer from server: $reply") unless $l eq "000001";
       $k = $l+6;
       &main::Die("Invalid answer from server: k=$k ") if $k > length($reply);
       $flavour = substr($reply,$[+6,$l) if $l;
       &main::Die("Invalid answer from server: $reply") unless $flavour =~ /[WL]/;
       &main::Log("krb flavour" . _refmt($flavour)) if $flavour && $main::Debug;
       $reply = substr($reply,$[+$k);      # now points to data part
    }

    return ($token, $krb5ticket, $reply, $flavour);
}

sub __Do_auth {
# call either __Do_token or __Do_kinit
my ($want_KRB, $token, $krb5ticket, $krb5file) = @_;
if ($want_KRB) {
   if ($krb5ticket) {
      unlink $krb5file;
      # if someone tries to out-race us, the sysopen in Do_kinit will fail
      &__Do_kinit(".", $krb5ticket, $krb5file) if ($want_KRB && $krb5ticket);
   } else {
      &Die("no krb5 ticket returned by server");
   }
} else {
   if ($token) {
      &__Do_token($token);
   } else {
      &Die("no token returned by server");
   }
}
}

sub __Do_token {
    my ($token) = @_;
    # set token
#   my $tok_arg = ($main::Debug) ? "-v" : "";
    my $SETTOKEN = _locate("SetToken", "/usr/sue/etc", "/afs/usr/local/etc");
#   open(T,"|$SETTOKEN $tok_arg") || &main::Die("SetToken: $!");
    open(T,"|$SETTOKEN         ") || &main::Die("SetToken: $!");
    print T $token;
    close(T);
}

sub __Do_kinit {
    #
    # must run with $< = $> = user id
    #
    use Fcntl;

    # Switch to use MIT kinit as default - The Heimdal version 
    # on the SLC4 machines is not compatable with AD by default.
    my $local_kinit = ""; #$kinit;

    if (-x "/usr/bin/kinit") { 
	$local_kinit  = "/usr/bin/kinit";
    }
    elsif (-x "/usr/kerberos/bin/kinit") {
	$local_kinit  = "/usr/kerberos/bin/kinit";
    }
    else
    {
	&Die("do not know where to look for kinit.\n");
    }


    #if ( -d "/usr/kerberos/bin" ) {
    #    $local_kinit  = "/usr/kerberos/bin/kinit";
    #}
    #elsif ( -e "/usr/bin/kinit" ) {
    #    $local_kinit  = "/usr/bin/kinit";
    #}

    my ($dummy, $krb5ticket, $krb5file) = (@_, "", "");
    if ($krb5file) {
      sysopen(F, "$krb5file", O_WRONLY|O_CREAT|O_EXCL, 0600)
           || &main::Die("Unable to open ticket file $krb5file for output");
      syswrite(F, $krb5ticket, length($krb5ticket));
      close(F);
    }
    while (1) {
       my $errmsg = `$local_kinit -R -f 2>&1`;


       # workaround for batch problem
       #if ( -d "/usr/kerberos/bin" ) {       
       #   my $errormsge2 = `/usr/kerberos/bin/kvno host/lxkerb03.cern.ch 2>&1`;
       #}
       #elsif ( -e "/usr/bin/kvno" ) {
       #   my $errormsge2 = `/usr/bin/kvno host/lxkerb03.cern.ch 2>&1`;
       #}

       my $rc = $?;
       if ($rc) {
         if ( $errmsg =~ /ASN\.1 value too large/ ) { sleep(2); next; }
         &main::Die("Cannot refresh ticket. rc from kinit: $rc, message is: $errmsg");
       } else {
         $errmsg = "" if $errmsg =~ /ticket lifetime is/;  # other NOTICE: messages are sent
         &main::Print("kinit returned \"$errmsg\"") if $errmsg;
       }
       last;
    }
#   (my $k524 = $local_kinit) =~ s/kinit/krb524init/;
#   system "$k524 ";
#   &main::Print("error in $k524: $!") if $?;
    my $aklog = ( -x "/usr/bin/aklog" ) ? "/usr/bin/aklog" : "/usr/bin/afs5log";
    system "$aklog";
    &main::Die("error in $aklog: $!") if $?;
}

sub _refmt {
    my($v) = @_;
    my $l = length($v);
    
    if ( $l > 16 ) {  $v = substr($v,$[,16); }
    $v =~ s/([\W])/sprintf("\\%02lx", ord $1)/eg;
    if ( $l > 16 ) {  $v .= "..."; }
    return " (length=$l)= $v";
}



sub _locate {
    my $survive = 0;
    my $cmd = shift(@_);
    if ($cmd eq "+") {  $survive=1; $cmd = shift(@_); }
    my $path = "";
    foreach my $d ((@_)) {
        if (-e "$d/$cmd") { $path = "$d/$cmd"; last; }
    }
    return $path if $path;
    &main::Die("cannot find $cmd") unless $survive;
    return "";
}


#------------------------------------------------------------------------------------------
#
# protect a few userids:
#   - members of system:administrators
#   - members of afs:vips
#   - system accounts, which should not have an entry in the pts or kdc databases
#
# returns (rc, text) in case of error or if user protected
#   - 1 :   user protected
#   - 2 :   missing userid
#   - 5 :   error opening file of protected users
#
#------------------------------------------------------------------------------------------
sub userid_exclude {
   my $file = "/usr/sue/etc/afs.reserved";
   my $pts_cmd = -r '/usr/afs/bin/pts' ? '/usr/afs/bin/pts' : '/usr/bin/pts';

   my($user) = @_;
   my(@ret) = caller(0);
   my @err;

   while(1) {
   @err = ( 2, "Missing userid"), last if $user eq '';
   @err = ( 1, "Please do not specify realm in $user"), last if $user =~ /\@/;
   @err = ( 1, "Character $1 refused in $user"), last if $user =~ /([^\w\-\+\/\.])/;
   @err = ( 1, "Refused userid $user"), last if $user =~ /\/admin/;
   $user = "\L$user\E";

   if ($user !~ /:/) {
     my(@res) = `$pts_cmd mem $user 2>/dev/null`;
     if (@res) {
       my @exclude = ( "system:administrators", "afs:vips" );
       @err = ( 1, "Protected or invalid userid $user"), last 
           if ( grep { my $a=$_; grep {$_ =~ /\s$a\s/} @res } @exclude );
     }
   }

   unless (open (F,"<$file")) { @err = ( 3, "Cannot open file $file"); last};
   my (@excluded) = <F>; 
   close(F);
   chomp(@excluded);
   @err = ( 1, "Protected or invalid userid $user"), last if ( grep {$_ eq $user} @excluded) ;
   @err = (); last;
   }
   if (@err) {
     if ($ret[5] eq "") {
        die "$err[1]\n";
     } else {
        return @err;
     }
   } else {
     return 0;
   }

}



#
# general subroutines
#

sub get_realm {
    my $k5_conf = "/etc/krb5.conf";
    my $realm;
    my @lines = `/bin/grep default_realm $k5_conf 2>/dev/null`;
    foreach (@lines) {
        ($realm) = $_ =~ /=\s+(\S+)\s*$/;
        return $realm if $realm;
    }
    $realm = "CERN.CH" unless $realm;      # we need some default
    return $realm;
}

sub get_kdc {
    my $k5_conf = "/etc/krb5.conf";
    my $myrealm = &get_realm;
    my $domain = lc($myrealm);
    my(@kdc) = ();
    my $lines = `/bin/grep kdc $k5_conf 2>/dev/null`;
    if ($lines) {
        foreach ( $lines =~ /=\s+(.+)\s*\n/go ) {
            foreach ( split /\s+/ ) {
                next unless /\.\Q$domain\E$/;
                push @kdc, $_;
            }
        }
        return (@kdc) if $#kdc;
    } 

    @kdc = ("afsdb1");        # needs some default there
    return @kdc;
}


sub llog {
   if (defined(&main::Log)) {
      &main::Log(@_);
   } elsif (defined(&main::log)) {
      &main::log(@_);
   } else {
      system("$logger", "--", "$$: @_");
   }
}

END {
}

1;
__END__

=head1 NAME

Heimtools - Perl interface to provide some kerberos5 additional tools

=head1 SYNOPSIS

  use Heimtools;
  setpag [(<dir>)];
  $x = authenticate (<principal> [, <keytab> ] )
  @l = get_users;
  block_unblock (<userid> [, <action> [, <reason> ] ] )
  new_password (<principal>, <pw string> [, <lifetime>] )
  $s = set_pwstring (<principal>, <password> [, <text prefix>] )
  expire_password (<principal>, <lifetime>)
  $s = set_expstring (<lifetime>)

=head1 DESCRIPTION

This package provides a few extra functions which help converting some AFS tools
to an Heimdal (kerberos5) environment.

=over 4

=item B<setpag>

	 setpag[(,dir>);

Creates the $KRB5CCNAME environment variable.
It assumes that an AFS PAG has already been created, and builds for the script and all its children
environment variables similar to those an Heimdal setpag command would have done.
The advantage is that the pagsh should be inserted BEFORE calling the script, which is not
always easy.

if called without argument, it derives the name of the ticket files from the AFS PAG identity.

if called with an argument, which must be a directory, it assumes this directory permanently contains
the ticket files, allowing subsequent calls to the script to reuse the same tickets, reducing the
number of calls to kinit.


=item B<authenticate>

	 authenticate(<principal> [, <keytab> ] );

Does the necessary kinit for the specified principal to setup the kerberos tickets.
If a keytab is specified, it uses it, else it uses the system /etc/krb5.keytab.

returns 1 in case of success, else returns 0.


=item B<get_users>

	 get_users;

reads the Heimdal database, extract user information and fills a few arrays:
   %is_user      :   1 if user exists in DB, otherwise not set
   %modified_on  :   date entry last modified (Heimdal format: yyyymmddHHMMSS)
   %modified_by  :   principal which did the modification
   %pw_expires   :   date password will expire / has expired (Heimdal format)
   %pw_expired   :   1 if password expired, otherwise not set
   %expires      :   date entry will expire / had expired (Heimdal format)
   %expired      :   1 if entry expired, otherwise not set


=item B<block_unblock>

	 block_unblock (<userid> [, <action> [, <reason> ] ] )

block/unblock an AFS account, optionally print the result

  <userid>       : the userid to be blocked or unblocked
  <action>       : 1 to block, 0 to unblock (defaults to unblock)
  <reason>       : text of the message to be printed
                   (it text set, also prints the result)

When a userid is blocked, his/her home directory is made unavailable.
Access is restablished as it was originally when account unblocked.

# Example :
# block_unblock('olemoign', 1) -> block account olemoign
# block_unblock('olemoign', 0) -> unblock account olemoign


=item B<new_password>

         new_password(<principal>, <pw string> [, <lifetime>] )

forces a new password for an account

  <principal>    : the principal to which the pw= string will apply
  <pw string>    : a string as returned by set_pwstring (see below)
  <lifetime>     : an optional lifetime for the password being set

runs a kadmin passwd command, locally or remotely, to set the new password
if a lifetime is specified, calls expire_password to set the expiration date of
the just set password.


=item B<set_pwstring>

         set_pwstring(<principal>, <password> [, <text prefix>] )

prepares the pw= argument for kadmin commands

  <principal>    : the principal to which the pw= string will apply
  <password>     : one of :
                   -r          : to ask for a random password
                   -           : to read password from STDIN
                   string      : the password itself
  <text prefix>  : optional string to be prepended to messages

returns either --random-key or --password=...... (for standard user)
or --key=...... if <user> is rcmd.<host> 


=item B<expire_password>

         expire_password(<principal>, <lifetime>)

forces a life time for the principal's current password

  <principal>    : the principal of which the password lifetime will be set
  <lifetime>     : the lifetime to be set for the current password. 

runs a kadmin modify command, locally or remotely, to set the new expiration date of
the current password.

Specify <lifetime> as "never" or a number of days


=item B<set_expstring>

         set_expstring(<lifetime>)

translates the specified password lifetime into an expiration date.

  <lifetime>     : the lifetime to be set for the current password. 

Specify <lifetime> as "never" or a number of days.
If the lifetime is specified as a number of days, that number is added to the current date, and the resulting
date returned. 
A negative number will make the password to be immediately expired.


=item B<userid_exclude>

         userid_exclude (<principal>)

verifies that <principal> is a syntacticly legal string, and not part of a list of "special" names.
This list contains service names appearing in /etc/passwd (e.g. root), and the accounts of
AFS administrators.

if the principal is acceptable, returns 0
else returns an error code and an explanatory text.


=item B<Krun>

         Krun ( <command list> )

runs the kadmin command created by amalgamating all the elements of the <list>.
each element of the list is first analyzed, then accordingly quoted, to avoid the danger presented by
strange characters.

A global variable $K_debug, set by the caller script, decides if the command is to be ran, 
or just printed as "whould run ......" .

To avoid problems due to concurrent invocations of scripts using Krun, a lock is taken while the command
runs.

B<future developments>

- When the caller script runs as root on the master kerberos server,
the kadmin command is transformed into B<kadmin -l>.

- IF the command doesnot use output redirection, run it directly from perl, without going through a shell

=back

=head1 AUTHORS

Bernard Antoine <Bernard.Antoine@cern.ch>

=head1 SEE ALSO

Volset and Vos modules

=cut
